1 VERSIONINFO
FILEVERSION     0,8,6,0
PRODUCTVERSION  0,8,6,0
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904B0"
    BEGIN
      VALUE "CompanyName", "davidgf.net"
      VALUE "FileDescription", "Whatsapp plugin for libpurple (pidgin)"
      VALUE "FileVersion", "0.8.6"
      VALUE "InternalName", "whatsapp4pidgin"
      VALUE "LegalCopyright", "Copyright 2015 David Guillen Fandos"
      VALUE "OriginalFilename", "libwhatsapp.dll"
      VALUE "ProductName", "Whatsapp for Purple"
      VALUE "ProductVersion", "0.8.6"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x409, 1200
  END
END
