#ifndef __WA_CONSTANTS__H__
#define __WA_CONSTANTS__H__

#define WHATSAPP_SERVER          "s.whatsapp.net"

#define WHATSAPP_VERSION         "Android-2.12.176-443"
#define WHATSAPP_USER_AGENT      "User-Agent: WhatsApp/2.12.176 Android/4.3 Device/GalaxyS3\r\n"

#define WHATSAPP_STATUS_ONLINE   "online"
#define WHATSAPP_STATUS_AWAY     "away"
#define WHATSAPP_STATUS_OFFLINE  "offline"

#define WHATSAPP_DEFAULT_PORT   443

#endif 
